/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | Copyright (C) 2007 Takuya OSHIMA
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    OpenFOAM is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the
    Free Software Foundation; either version 2 of the License, or (at your
    option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM; if not, write to the Free Software Foundation,
    Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA

Description
    Reads probed time series of a scalar variable and outputs as a
    .wav file or send to /dev/dsp.

Author
    Takuya OSHIMA <oshima@eng.niigata-u.ac.jp>

\*---------------------------------------------------------------------------*/

#if defined(linux) || defined(linux64) || defined(linuxAMD64) || defined(cygwin)
#define PLAY_SOUND 1
#define USE_ALSA 0
#endif

#include "argList.H"
#include "DynamicList.H"
#include "IStringStream.H"
#ifdef cygwin
#include "Time.hh"
#else
#include "Time.H"
#endif

#include <fstream>
#include <sstream>

#if PLAY_SOUND
#if USE_ALSA
#include <alsa/asoundlib.h>
#else
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/soundcard.h>
#include <sys/stat.h>
#include <unistd.h>
#endif
#endif

using namespace Foam;

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

bool readSeries(DynamicList<scalar>& series, const fileName& probeName,
const label column, const scalar startTime)
{
    std::ifstream in(std::string(probeName).c_str());

    if(!in.good())
    {
        FatalError << "Can't open " << probeName << exit(FatalError);
    }

    std::string line;
    std::getline(in, line);

    // probe file created by OF 1.3 begins with "Time", whereas >= 1.4 "#   "
    if(line.length() < 4
        || (line.substr(0, 4) != "#   " && line.substr(0, 4) != "Time"))
    {
        FatalError << "The first line of " << probeName << " is invalid"
            << exit(FatalError);
    }

    // count the number of columns
    label nColumns = 1;
    forAll(line, charI)
    {
        if(line[charI] == '(')
        {
            nColumns++;
        }
    }
    if(column < 1 || column >= nColumns)
    {
        FatalError << "Tried to read column " << column
            << " while the file only contains " << nColumns << " columns"
            << exit(FatalError);
    }

    // read values
    bool showedStartTime = false;
    while(!in.eof())
    {
        std::getline(in, line);
        std::istringstream iss(line);

        // read time
        scalar currentTime;
        iss >> currentTime;

        if(currentTime >= startTime)
        {
            if(!showedStartTime)
            {
                Info << "Starting conversion from Time = " << currentTime
                    << endl << endl;
                showedStartTime = true;
            }

            // skip until the specified column
            for(label columnI = 1; columnI < column; ++columnI)
            {
                std::string dummyString;
                iss >> dummyString;
            }

            // read the column
            scalar scalarValue;
            iss >> scalarValue;
            series.append(scalarValue);
        }
    }
    in.close();
    series.shrink();
    return true;
}

void seriesToPcmData(List<int16_t>& data, const DynamicList<scalar>& series)
{
    data.setSize(series.size());

    // find the min and max sample values and calculate bias
    scalar minValue = HUGE_VAL, maxValue = -HUGE_VAL, sum = 0.0;
    forAll(series, sampleI)
    {
        minValue = min(minValue, series[sampleI]);
        maxValue = max(maxValue, series[sampleI]);
        sum += series[sampleI];
    }
    const scalar bias = sum / static_cast<scalar>(series.size());
    Info << "Normalizing recording level:" << endl;
    Info << "    Min value = " << minValue << endl;
    Info << "    Max value = " << maxValue << endl;
    Info << "    Bias = " << bias << endl;

    // scale and convert the data to be fit into the int16_t range
    scalar scaleFactor = max(bias - minValue, maxValue - bias);
    if(fabs(scaleFactor) < VSMALL)
    {
        Warning << "Wave amplitude is too small to be normalized." << endl
            << "The converted data will not be normalized." << endl;
        scaleFactor = 1.0;
    }
    scalar squaredSum = 0.0;
    forAll(data, sampleI)
    {
        data[sampleI] = static_cast<short>((series[sampleI] - bias)
            / scaleFactor * 32767.0);
        squaredSum += static_cast<scalar>(data[sampleI] * data[sampleI]);
    }
    const scalar level = 10.0 * Foam::log10(squaredSum
        / (static_cast<scalar>(data.size()) * 32767.0 * 32767.0));
    Info << "    Average level = " << level << " dB" << endl << endl;
}

bool pcmDataToWav(const fileName& wavName, const List<int16_t>& data,
    const uint32_t samplingRate)
{
    // open the file
    std::ofstream out(std::string(wavName).c_str());
    if(!out.good())
    {
        FatalError << "Can't open " << wavName << exit(FatalError);
    }

    // calculate the header information
    const uint32_t fmtChunkSize = 16;
    const uint16_t formatId = 1;
    const uint16_t nChannels = 1;
    const uint16_t bitsPerSample = 16;
    const uint32_t dataRate = samplingRate * bitsPerSample / 8;
    const uint16_t blockSize = bitsPerSample / 8 * nChannels;
    const uint32_t dataSize = data.size() * 2;
    const uint32_t fileSizeMinus8 = dataSize + 36;

    // write header
    out.write("RIFF", 4);
    out.write(reinterpret_cast<const char *>(&fileSizeMinus8), 4);
    out.write("WAVE", 4);
    out.write("fmt ", 4);
    out.write(reinterpret_cast<const char *>(&fmtChunkSize), 4);
    out.write(reinterpret_cast<const char *>(&formatId), 2);
    out.write(reinterpret_cast<const char *>(&nChannels), 2);
    out.write(reinterpret_cast<const char *>(&samplingRate), 4);
    out.write(reinterpret_cast<const char *>(&dataRate), 4);
    out.write(reinterpret_cast<const char *>(&blockSize), 2);
    out.write(reinterpret_cast<const char *>(&bitsPerSample), 2);
    out.write("data", 4);
    out.write(reinterpret_cast<const char *>(&dataSize), 4);

    // write body
    out.write(reinterpret_cast<const char *>(&data[0]), dataSize);

    out.close();
    return true;
}

#if PLAY_SOUND
bool pcmDataToDevice(const List<int16_t>& data, uint32_t samplingRate)
{
#if USE_ALSA
    const char *pcm_name = "plughw:0,0";

    // open the audio device
    snd_pcm_t *pcm_handle;
    if(snd_pcm_open(&pcm_handle, pcm_name, SND_PCM_STREAM_PLAYBACK, 0) < 0)
    {
        FatalError << "Can't open PCM device" << pcm_name << exit(FatalError);
    }

    // setup the audio device
    snd_pcm_hw_params_t *hwparams;
    if (snd_pcm_hw_params_malloc (&hwparams) < 0)
    {
        snd_pcm_close(pcm_handle);
        FatalError << "Can't allocate hardware parameter structure" 
            << exit(FatalError);
    }
    if(snd_pcm_hw_params_any(pcm_handle, hwparams) < 0)
    {
        snd_pcm_close(pcm_handle);
        FatalError << "Can not configure this PCM device" << exit(FatalError);
    }
    if(snd_pcm_hw_params_set_access(pcm_handle, hwparams,
        SND_PCM_ACCESS_RW_INTERLEAVED) < 0)
    {
        snd_pcm_close(pcm_handle);
        FatalError << "Error setting access" << exit(FatalError);
    }
    if (snd_pcm_hw_params_set_format(pcm_handle, hwparams,
        SND_PCM_FORMAT_S16_LE) < 0)
    {
        snd_pcm_close(pcm_handle);
        FatalError << "Error setting format" << exit(FatalError);
    }
    unsigned int exact_rate = samplingRate;
    if(snd_pcm_hw_params_set_rate_near(pcm_handle, hwparams, &exact_rate, 0)
        < 0)
    {
        snd_pcm_close(pcm_handle);
        FatalError << "Error setting rate" << exit(FatalError);
    }
    if(exact_rate != samplingRate)
    {
        Warning << "The sampling rate " << samplingRate
            << " Hz is not supported by your hardware."
            << " Using " << exact_rate << " Hz instead." << endl;
    }
    if (snd_pcm_hw_params_set_channels(pcm_handle, hwparams, 1) < 0)
    {
        snd_pcm_close(pcm_handle);
        FatalError << "Error setting channels." << exit(FatalError);
    }
    if (snd_pcm_hw_params(pcm_handle, hwparams) < 0)
    {
        snd_pcm_close(pcm_handle);
        FatalError << "Error setting HW params" << exit(FatalError);
    }
    snd_pcm_hw_params_free(hwparams);

    // write the PCM data to the device
    if (snd_pcm_writei(pcm_handle, &data[0], data.size()) != data.size())
    {
        snd_pcm_close(pcm_handle);
        FatalError << "Write to audio interface failed" << exit(FatalError);
    }

    // close the device
    snd_pcm_close(pcm_handle);

    return true;
#else
    // ioctl settings
    const std::string deviceName = "/dev/dsp";
    const int format = AFMT_S16_LE;
    const int nChannels = 1;

    // open the audio device
    const int fd = open(deviceName.c_str(), O_WRONLY);
    if(fd == -1)
    {
        FatalError << "Can't open " << deviceName
            << ". Check if you have write permission." << exit(FatalError);
    }

    // setup the audio device
    if(ioctl(fd, SOUND_PCM_SETFMT, &format) == -1)
    {
        close(fd);
        FatalError << "Can't set format to PCM" << exit(FatalError);
    }
    if(ioctl(fd, SOUND_PCM_WRITE_CHANNELS, &nChannels) == -1)
    {
        close(fd);
        FatalError << "Can't set the number of channels to " << nChannels
            << exit(FatalError);
    }
    const int rate = static_cast<int>(samplingRate);
    if(ioctl(fd, SOUND_PCM_WRITE_RATE, &rate) == -1)
    {
        close(fd);
        FatalError << "Can't set the sampling rate to " << rate
            << exit(FatalError);
    }

    // write the PCM data to the device
    write(fd, &data[0], data.size() * sizeof(int16_t));

    // close the device
    close(fd);

    return true;
#endif
}
#endif

// Main program:

int main(int argc, char *argv[])
{
    argList::noParallel();
    argList::validArgs.append("probe file");
    argList::validOptions.insert("column",
        "<the column number to be converted in probe file"
        " (starts from 0; default 1)>");
#if PLAY_SOUND
    argList::validOptions.insert("play", "");
#endif
    argList::validOptions.insert("startTime",
        "<startTime (default -HUGE_VAL)>");
    argList::validOptions.insert("timeStretchRatio",
        "<time stretch ratio (default 1.0)>");

#include "setRootCase.H"

    Info << endl;

#   include "createTime.H"

    fileName probeName(args.args()[3]);

    label column = 1;
    if(args.options().found("column"))
    {
        column = readLabel(IStringStream(args.options()["column"])());
    }

    scalar startTime = -HUGE_VAL;
    if(args.options().found("startTime"))
    {
        startTime = readScalar(IStringStream(args.options()["startTime"])());
    }

    Info << "Converting the column " << column << " of " << probeName
        << " to 16-bit monaural linear PCM wave data" << endl << endl;

    // read the probe file
    DynamicList<scalar> series;
    readSeries(series, probeName, column, startTime);

    scalar timeStretchRatio = 1.0;
    if(args.options().found("timeStretchRatio"))
    {
        timeStretchRatio = readScalar(IStringStream(
            args.options()["timeStretchRatio"])());
    }

    // convert floting point values to integers
    List<int16_t> data;
    seriesToPcmData(data, series);

    // calculate the sampling rate
    const uint32_t samplingRate = static_cast<uint32_t>(round(1.0
        / (runTime.deltaT().value() * timeStretchRatio)));

    Info << "PCM wave data info:" << endl;
    Info << "    Sampling rate = " << static_cast<int>(samplingRate) << " Hz"
        << endl;
    Info << "    Number of samples = " << data.size() << endl;
    Info << "    Duration = " << static_cast<scalar>(data.size())
        / static_cast<scalar>(samplingRate) << " s" << endl;
    if(samplingRate < 1000)
    {
        Warning << "Sampling rate is below 1 kHz." << endl
            << "The converted data may not be played correctly on some"
            << " audio devices." << endl;
    }
    else if(samplingRate > 48000)
    {
        Warning << "Sampling rate exceeds 48 kHz." << endl
            << "The converted data may not be played correctly on some"
            << " audio devices." << endl;
    }

    // write the data to a file or a device
#if PLAY_SOUND
    if(args.options().found("play"))
    {
        Info << endl << "Playing the converted data" << endl;
        pcmDataToDevice(data, samplingRate);
    }
    else
    {
#endif
        const fileName wavName = probeName.lessExt() + "_" + name(column)
            + ".wav";

        Info << endl << "Writing the converted data to " << wavName << endl;
        pcmDataToWav(wavName, data, samplingRate);
#if PLAY_SOUND
    }
#endif

    Info << endl << "End\n" << endl;

    return 0;
}


// ************************************************************************* //
